# $Id: datagathering.tcl 1572 2008-10-22 09:35:56Z sergei $
#
# Data Forms (XEP-0004) support
#

namespace eval data {
    set winid 0

    # Registration & search fields (see XEP-0077 & XEP-0055)

    array set field_labels [list \
	username     [::msgcat::mc "Username:"] \
	nick         [::msgcat::mc "Nickname:"] \
	password     [::msgcat::mc "Password:"] \
	name         [::msgcat::mc "Full Name:"] \
	first        [::msgcat::mc "First Name:"] \
	last         [::msgcat::mc "Last Name:"] \
	email        [::msgcat::mc "E-mail:"] \
	address      [::msgcat::mc "Address:"] \
	city         [::msgcat::mc "City:"] \
	state        [::msgcat::mc "State:"] \
	zip          [::msgcat::mc "Zip:"] \
	phone        [::msgcat::mc "Phone:"] \
	url          [::msgcat::mc "URL:"] \
	date         [::msgcat::mc "Date:"] \
	misc         [::msgcat::mc "Misc:"] \
	text         [::msgcat::mc "Text:"] \
	key          [::msgcat::mc "Key:"]]

    disco::register_feature jabber:x:data
}

proc data::fill_fields {g items} {
    variable data
    variable field_labels

    set row 0
    set data(varlist,$g) {}
    
    grid columnconfig $g 1 -weight 1 -minsize 0

    foreach item $items {
	::xmpp::xml::split $item tag xmlns attrs cdata subels
	if {$xmlns == "jabber:x:data" || $xmlns == "jabber:iq:data"} {
	    return [fill_fields_x $g $subels]
	}
    }

    set focus ""
    set fields {}
    foreach item $items {
	::xmpp::xml::split $item tag xmlns attrs cdata subels

	switch -- $tag {
	    instructions {
		message $g.instructions$row -text $cdata -width 10c
		grid $g.instructions$row -row $row -column 0 -columnspan 2 \
		    -sticky w -pady 2m
		incr row
	    }
	    registered -
	    x {}
	    default {
                lappend fields $tag $cdata
	    }
	}
    }
    foreach {tag cdata} $fields {
	lappend data(varlist,$g) $tag

	if {[info exists field_labels($tag)]} {
	    label $g.l$row -text $field_labels($tag)
	} else {
	    label $g.l$row -text $tag
	}
	switch -- $tag {
	    key {
		entry $g.$row \
		    -textvariable [namespace current]::data(var,$tag,$g) \
		    -state disabled
	    }
	    password {
		entry $g.$row \
		    -textvariable [namespace current]::data(var,$tag,$g) \
		    -show *
		if {$focus == ""} {
		    set focus $g.$row
		}
	    }
	    default {
		entry $g.$row \
		    -textvariable [namespace current]::data(var,$tag,$g)
		if {$focus == ""} {
		    set focus $g.$row
		}
	    }
	}

	if {$cdata != {}} {
	    set data(var,$tag,$g) $cdata
	}

	grid $g.l$row -row $row -column 0 -sticky e
	grid $g.$row  -row $row -column 1 -sticky we

	incr row
    }
    return $focus
}

proc data::cleanup {g} {
    variable data

    array unset data *,$g
}

proc data::get_tags {g} {
    variable data

    if {[info exists data(x,$g)]} {
	return [get_tags_x $g]
    }

    set restags {}

    if {[info exists data(varlist,$g)]} {
	foreach var $data(varlist,$g) {
	    lappend restags [::xmpp::xml::create $var -cdata $data(var,$var,$g)]
	}
    }

    return $restags
}

proc data::get_reported_fields {g} {
    variable data

    return $data(varlist,$g)
}

###############################################################################
# x:data processing
###############################################################################

# create field tag

proc ::data::createfieldtag {type args}  {
    array set params $args

    switch -- $type {
	instructions -
	title {
	    if {[info exists params(-value)]} {
		return [::xmpp::xml::create $type -cdata $params(-value)]
	    } else {
		error "You must define -value"
	    }
	}

	fixed - 
	hidden -
	list-single -
	list-multi -
	text-single - 
	text-multi -
	text-private -
	jid-multi -
	jid-single -
	boolean {
	    set vars [list type $type]

	    if {[info exists params(-var)]} {
		lappend vars var $params(-var)
	    } elseif {![string equal $type fixed]} {
		error "You must define -var"
	    }

	    if {[info exists params(-label)]} {
		lappend vars label $params(-label)
	    }

	    set subtags  {}
	    if {[info exists params(-descr)]} {
		lappend subtags [::xmpp::xml::create descr \
				     -cdata $params(-descr)]
	    }
	    if {[info exists params(-required)] && $params(-required)} {
		lappend subtags [::xmpp::xml::create required]
	    }
	    if {[lcontain {jid-multi text-multi list-multi hidden fixed} $type]} {
		if {[info exists params(-values)]} {
		    foreach value $params(-values) {
			lappend subtags [::xmpp::xml::create value \
					     -cdata $value]
		    }
		} elseif {[lcontain {jid-multi hidden fixed} $type]} {
		    error "You must define -values"
		}
	    } else {	    
		if {[info exists params(-value)]} {
		    lappend subtags [::xmpp::xml::create value \
					 -cdata $params(-value)]
		}
	    }
		
	    if {[lcontain {list-multi list-single} $type]} {
		if {[info exists params(-options)]} {
		    foreach option $params(-options) {
			lassign $option name label
			lappend subtags \
			    [::xmpp::xml::create option \
				 -attrs [list label $label] \
				 -subelement [::xmpp::xml::create value \
							-cdata $name]]
		    }
		} else {
		    error "You must define -options"
		}
	    }
	    
	    return [::xmpp::xml::create field -attrs $vars \
					      -subelements $subtags]	    
	}
	
	default {
	    error "Unknown type $type"
	}
    }   
}

# parse_result
proc data::parse_xdata_results {items args} {
    set report_hidden 0
    foreach {key val} $args {
	switch -- $key {
	    -hidden { set report_hidden $val }
	}
    }

    set result {}
    foreach item $items {
	::xmpp::xml::split $item tag xmlns attrs cdata subels
	if {$tag != "field"} {
	    continue
	}
	set type [::xmpp::xml::getAttr $attrs type]
	if {!$report_hidden && $type == "hidden"} {
	    continue
	}
	set var [::xmpp::xml::getAttr $attrs var]
	if {$var == ""} {
	    continue
	}
	set label [::xmpp::xml::getAttr $attrs label]
	set values {}
	foreach subel $subels {
	    ::xmpp::xml::split $subel stag sxmlns sattrs scdata ssubels
	    if {$stag == "value"} {
		lappend values $scdata
	    }
	}
	lappend result [list $var $type $label $values]
    }
    return $result
}

proc data::add_label {g row label {required 0}} {
    if {$label != ""} {
	if {$required} {
	    set prefix *
	} else {
	    set prefix ""
	}
	if {![string is punct [string index $label end]]} {
	    set suffix :
	} else {
	    set suffix ""
	}
	label $g.label$row -text ${prefix}${label}$suffix
	grid $g.label$row -row $row -column 0 -sticky en
    }
}

proc data::render_media {g row media_list} {
    foreach item $media_list {
	::xmpp::xml::split $item tag xmlns attrs cdata subels
	set unsupported 1
	foreach subel $subels {
	    ::xmpp::xml::split $subel stag sxmlns sattrs scdata ssubels
	    switch -- $stag {
		uri {
		    render_url $g.mediauri$row $scdata $scdata -bg [get_conf $g -bg]
		    grid $g.mediauri$row -row $row -column 1 -sticky ew
		    set unsupported 0
		    incr row
		}
		data {
		    set type [::xmpp::xml::getAttr $sattrs type]
		    switch -glob -- $type {
			image/* {
			    if {![catch {image create photo -data [base64::decode $scdata]} img]} {
				label $g.mediaimg$row -image $img
				bind $g.mediaimg$row <Destroy> [list image delete $img]
				grid $g.mediaimg$row -row $row -column 1 -sticky ew
				set unsupported 0
				incr row
			    }
			}
			default {
			    # TODO: implement other types
			}
		    }
		}
	    }
	}
	if {$unsupported} {
	    # No supported media item
	    return -code error "No supported types for a media element"
	}
    }
    return $row
}

proc data::fill_fields_x {g items} {
    variable data

    set data(x,$g) 1
    set row 0
    set data(allvarlist,$g) {}
    set focus ""

    foreach item $items {
	::xmpp::xml::split $item tag xmlns attrs cdata subels

	switch -- $tag {
	    instructions {
		#add_label $g $row [::msgcat::mc "Instructions"]
		message $g.instructions$row -text $cdata -width 15c
		grid $g.instructions$row -row $row -column 0 \
		    -columnspan 2 -sticky w -pady 2m
		incr row
	    }
	    field {
		set widget [fill_field_x $g $row $tag $attrs $cdata $subels]
		if {$focus == ""} {
		    set focus $widget
		}
		incr row
	    }
	    title {
		set top [winfo toplevel $g]
		if {$top != "."} {
		    wm title $top $cdata
		    wm iconname $top $cdata
		}
	    }
	    default {
		debugmsg filetransfer "XDATA: unknown tag $tag"
	    }
	}
    }

    # FIX THIS
    set data(varlist,$g) $data(allvarlist,$g)

    return $focus
}


proc data::fill_field_x {g row tag attrs cdata subels} {
    variable data

    set required 0
    set desc {}
    set options {}
    set vals {}
    set var   [::xmpp::xml::getAttr $attrs var]
    set type  [::xmpp::xml::getAttr $attrs type]
    if {$type == ""} {
	set type text-single
    }
    set label [::xmpp::xml::getAttr $attrs label]
    set data(var,$var,$g) ""
    set widget ""
    set media_list {}

    foreach item $subels {
	::xmpp::xml::split $item stag sxmlns sattrs scdata ssubels
	switch -- $stag {
	    required {set required 1}
	    value {
		set data(var,$var,$g) $scdata
		lappend vals $scdata
	    }
	    desc {set desc $scdata}
	    option {
		set lab [::xmpp::xml::getAttr $sattrs label]
		foreach item $ssubels {
		    ::xmpp::xml::split $item \
			sstag ssxmlns ssattrs sscdata sssubels
		    switch -- $sstag {
			value {set val $sscdata}
		    }
		}
		lappend options $lab $val
	    }
	    media {
		if {$sxmlns == $::NS(media-element)} {
		    lappend media_list $item
		}
	    }
	}
    }

    switch -- $type {
	jid-single -
	text-single -
	text-private {
	    add_label $g $row $label $required
	    set row [render_media $g $row $media_list]
	    entry $g.entry$row \
		-textvariable [namespace current]::data(var,$var,$g)
	    if {$type == "text-private"} {
		$g.entry$row configure -show *
	    }
	    grid $g.entry$row  -row $row -column 1 -sticky we
	    set widget $g.entry$row
	    if {$desc != ""} {
		balloon::setup $g.entry$row -text $desc
	    }
	}
	jid-multi -
	text-multi {
	    add_label $g $row $label $required
	    set row [render_media $g $row $media_list]
	    set sw [ScrolledWindow $g.textsw$row -scrollbar vertical]
	    textUndoable $g.text$row -height 6 -width 50
	    $sw setwidget $g.text$row
	    bind $g.text$row <Control-Key-Return> { }
	    bind $g.text$row <Return> "[bind Text <Return>]\nbreak"
	    set data(var,$var,$g) [join $vals \n]
	    catch { $g.text$row insert end $data(var,$var,$g) }
	    grid $sw -row $row -column 1 -sticky we
	    set data(text,$var,$g) $g.text$row
	    set widget $g.text$row
	    if {$desc != ""} {
		balloon::setup $g.text$row -text $desc
	    }
	}
	boolean {
	    switch -- $data(var,$var,$g) {
		1 -
		0 {
		    set onvalue 1
		    set offvalue 0
		}
		true -
		false {
		    set onvalue true
		    set offvalue false
		}
		default {
		    set onvalue 1
		    set offvalue 0
		    set data(var,$var,$g) 0
		}
	    }
	    add_label $g $row $label $required
	    set row [render_media $g $row $media_list]
	    checkbutton $g.cb$row \
		-variable [namespace current]::data(var,$var,$g) \
		-onvalue $onvalue -offvalue $offvalue
	    grid $g.cb$row  -row $row -column 1 -sticky w
	    set widget $g.cb$row
	    if {$desc != ""} {
		balloon::setup $g.cb$row -text $desc
	    }
	}
	fixed {
	    add_label $g $row $label $required
	    set row [render_media $g $row $media_list]
	    catch { message $g.m$row -text [join $vals \n] -width 10c }
	    grid $g.m$row -row $row -column 1 -sticky w
	    set dont_report 1
	    if {$desc != ""} {
		balloon::setup $g.m$row -text $desc
	    }
	}
	list-single {
	    add_label $g $row $label $required
	    set row [render_media $g $row $media_list]
	    set height 0
	    foreach {lab val} $options {
		lappend data(combol$row,$var,$g) $lab
		incr height
		if {[string equal $data(var,$var,$g) $val]} {
		    set data(combov$row,$var,$g) $lab
		}
	    }
	    if {$height > 10} {
		set height 10
	    }
	    set cb [ComboBox $g.combo$row \
			-height $height \
			-editable no \
			-values $data(combol$row,$var,$g) \
			-textvariable \
			[namespace current]::data(combov$row,$var,$g)]
	    grid $cb -row $row -column 1 -sticky we
	    trace variable [namespace current]::data(combov$row,$var,$g) w \
		[list data::trace_combo $options \
		     [namespace current]::data(var,$var,$g)]
	    set widget $g.combo$row
	    if {$desc != ""} {
		balloon::setup $g.combo$row -text $desc
	    }
	}
	list-multi {
	    add_label $g $row $label $required
	    set row [render_media $g $row $media_list]
	    set sw [ScrolledWindow $g.sw$row]
	    set l [listbox $g.lb$row -height 6 \
		       -selectmode multiple -exportselection no]
	    $sw setwidget $l
	    foreach {lab val} $options {
		$l insert end $lab
		if {[lcontain $vals $val]} {
		    $l selection set end
		}
	    }
	    grid $sw  -row $row -column 1 -sticky we
	    set data(multi,$var,$g) 1
	    trace_listmulti $l $options \
		data::data(var,$var,$g)
	    bind $l <<ListboxSelect>> \
		[list data::trace_listmulti %W $options \
		     [namespace current]::data(var,$var,$g)]
	    set widget $sw
	    if {$desc != ""} {
		balloon::setup $g.lb$row -text $desc
	    }
	}
	hidden {}

	default {
	    debugmsg filetransfer "XDATA: unknown field type '$type'"
	}
    }

    if {![info exists dont_report]} {
	lappend data(allvarlist,$g) $var
    }
    return $widget
}

proc data::trace_combo {assoc dst name1 name2 op} {
    foreach {lab val} $assoc {
	if {[string equal $lab [set ${name1}($name2)]]} {
	    set $dst $val
	}
    }
}

proc data::trace_listmulti {l assoc dst} {
    set $dst {}
    foreach idx [$l curselection] {
	#debugmsg filetransfer [lindex $assoc [expr $idx * 2 + 1]]
	lappend $dst [lindex $assoc [expr $idx * 2 + 1]]
    }
}


proc data::get_tags_x {g} {
    variable data

    set restags {}

    foreach var $data(varlist,$g) {
	if {[info exists data(multi,$var,$g)]} {
	    set vartags {}
	    foreach val $data(var,$var,$g) {
	        lappend vartags [::xmpp::xml::create value -cdata $val]
	    }
	} elseif {[info exists data(text,$var,$g)]} {
	    set data(var,$var,$g) [$data(text,$var,$g) get 1.0 "end -1c"]
	    set vartags {}
	    foreach val [split $data(var,$var,$g) \n] {
	        lappend vartags [::xmpp::xml::create value -cdata $val]
	    }
	} else {
	    set vartags [list [::xmpp::xml::create value \
					    -cdata $data(var,$var,$g)]]
	}
	lappend restags [::xmpp::xml::create field \
					     -attrs [list var $var] \
					     -subelements $vartags]
    }

    set restag [list [::xmpp::xml::create x \
					  -xmlns jabber:x:data \
					  -attrs [list type submit] \
					  -subelements $restags]]

    return $restag
}

###############################################################################

proc data::draw_window {items send_cmd {cancel_cmd destroy}} {
    variable winid

    set w .datagathering$winid
    incr winid

    if {[winfo exists $w]} {
	destroy $w
    }

    toplevel $w -class XData
    wm group $w .
    wm title $w ""
    wm iconname $w ""
    wm transient $w .
    wm withdraw $w
    set geometry [option get $w geometry XData]
    if {$geometry != ""} {
	wm geometry $w $geometry
    }

    set sw [ScrolledWindow $w.sw]
    set sf [ScrollableFrame $w.fields -constrainedwidth yes]
    set f [$sf getframe]
    $sf configure -height 10
    $sw setwidget $sf
    if {[catch {data::fill_fields $f $items} focus]} {
	destroy $w
	return -code error $focus
    }

    set bbox [ButtonBox $w.bbox -spacing 10 -padx 10 -default 0]
    pack $bbox -side bottom -anchor e -padx 2m -pady 2m
    $bbox add -text [::msgcat::mc "Send"] \
	-command [list eval $send_cmd [list $w] \[data::get_tags $f\]]
    $bbox add -text [::msgcat::mc "Cancel"] \
	-command [list eval $cancel_cmd [list $w]]
    bind $w <Return> [list ButtonBox::invoke $bbox default]
    bind $w <Escape> [list ButtonBox::invoke $bbox 1]
    bind $f <Destroy> [list [namespace current]::cleanup $f]

    bindscroll $f $sf

    pack [Separator $w.sep] -side bottom -fill x  -pady 1m

    set hf [frame $w.error]
    pack $hf -side top
    set vf [frame $w.vf]
    pack $vf -side left -pady 2m
    pack $sw -side top -expand yes -fill both -padx 2m -pady 2m

    update idletasks
    $hf configure -width [expr {[winfo reqwidth $f] + [winfo pixels $f 1c]}]

    set h [winfo reqheight $f]
    set sh [winfo screenheight $w]
    if {$h > $sh - 200} {
	set h [expr {$sh - 200}]
    }
    $vf configure -height $h
    wm deiconify $w
    if {$focus != ""} {
	focus $focus
    }

    return $w
}

###############################################################################

proc data::request_data {xmlns xlib jid node args} {
    if {$node == ""} {
	set vars {}
    } else {
	set vars [list node $node]
    }

    ::xmpp::sendIQ $xlib get \
	-query [::xmpp::xml::create query \
				    -xmlns $xmlns \
				    -attrs $vars] \
	-to $jid \
	-command [list [namespace current]::receive_data $xlib $xmlns $jid $node]
}

proc data::receive_data {xlib xmlns jid node res child} {
    if {[string equal $res abort]} {
	return
    }

    if {[string equal $res error]} {
	set ew .data_err
	if {[winfo exists $ew]} {
	    destroy $ew
	}
	MessageDlg $ew -aspect 50000 -icon error \
	    -message [::msgcat::mc "Error requesting data: %s" \
				   [error_to_string $child]] \
	    -type user -buttons ok -default 0 -cancel 0
	return
    }

    switch -- $xmlns {
	jabber:iq:data {
	    set subels [list $child]
	}
	default {
	    ::xmpp::xml::split $child tag xmlns attrs cdata subels
	}
    }

    data::draw_window $subels \
	[list [namespace current]::send_data $xlib $xmlns $jid $node] \
	[list [namespace current]::cancel_data $xlib $xmlns $jid $node]
}

proc data::cancel_data {xlib xmlns jid node w} {
    send_data $xlib $xmlns $jid $node $w \
	      [list [::xmpp::xml::create x \
			    -xmlns jabber:x:data \
			    -attrs [list type cancel]]]
}

proc data::send_data {xlib xmlns jid node w restags} {
    switch -- $xmlns {
	jabber:iq:data {
	    ::xmpp::xml::split [lindex $restags 0] tag xmlns1 attrs cdata subels
	}
	default {
	    set subels $restags
	    set attrs {}
	}
    }

    if {$node != ""} {
        lappend attrs node $node
    }

    destroy $w.error.msg
    $w.bbox itemconfigure 0 -state disabled

    ::xmpp::sendIQ $xlib set \
	    -query [::xmpp::xml::create query \
					-xmlns $xmlns \
					-attrs $attrs \
					-subelements $subels] \
	    -to $jid \
	    -command [list [namespace current]::test_error_res $w]
}

proc data::test_error_res {w res child} {
    if {[string equal $res ok]} {
	destroy $w
	return
    }

    $sw.bbox itemconfigure 0 -state normal

    set m [message $w.error.msg \
		   -aspect 50000 \
		   -text [error_to_string $child] \
		   -pady 2m]
    $m configure -foreground [option get $m errorForeground Message]
    pack $m
}

disco::browser::register_feature_handler jabber:iq:data \
    [list [namespace current]::data::request_data jabber:iq:data] -node 1 \
    -desc [list * [::msgcat::mc "Data form"]]

disco::browser::register_feature_handler ejabberd:config \
    [list [namespace current]::data::request_data ejabberd:config] -node 1 \
    -desc [list * [::msgcat::mc "Configure service"]]

