# $Id: negotiate.tcl 1520 2008-10-17 17:16:09Z sergei $

namespace eval negotiate {
    set ::NS(negotiate) http://jabber.org/protocol/feature-neg
    set seq 0
}

proc negotiate::get_handler {type xlib from child args} {
    variable handler

    debugmsg negotiate "$type: [list $from $child]"

    ::xmpp::xml::split $child tag xmlns attrs cdata subels

    set lang [::xmpp::xml::getAttr $args -lang en]

    set error 1
    set fields {}
    foreach form $subels {
	::xmpp::xml::split $form stag sxmlns sattrs scdata ssubels
	if {$stag == "x" && \
		$xmlns == $::NS(data) && \
		[::xmpp::xml::getAttr $sattrs type] == "submit"} {

		foreach field $ssubels {
		    ::xmpp::xml::split $field \
			    sstag ssxmlns ssattrs sscdata sssubels

		    if {$sstag != "field"} continue

		    set feature [::xmpp::xml::getAttr $ssattrs var]

		    if {![info exists handler($feature)]} continue

		    set error 0

		    # TODO
		    set opts [eval $handler($feature) \
				   [list $type $xlib $from $lang $sssubels]]

		    lappend fields $opts
	    }
	}
    }
    if {$error} {
	return [list error cancel feature-not-implemented]
    } else {
	set res [::xmpp::xml::create feature \
		     -xmlns $::NS(negotiate) \
		     -subelement [::xmpp::xml::create x \
					 -xmlns $::NS(data) \
					 -attrs [list type result] \
					 -subelements $fields]]
	return [list result $res]
    }
}

::xmpp::iq::register get feature $::NS(negotiate) \
		     [list [namespace current]::negotiate::get_handler get]
::xmpp::iq::register set feature $::NS(negotiate) \
		     [list [namespace current]::negotiate::get_handler set]

proc negotiate::register_handler {feature h} {
    variable handler

    set handler($feature) $h
}

proc negotiate::send_request {xlib to feature} {
    variable seq
    variable tmp

    set i [incr seq]

    set fieldtags {}
    if {$feature != ""} {
	lappend fieldtags [::xmpp::xml::create field \
				-attrs [list var $feature]]
    }

    ::xmpp::sendIQ $xlib get \
	-query [::xmpp::xml::create feature \
			-xmlns $::NS(negotiate) \
			-subelement [::xmpp::xml::create x \
					    -xmlns $::NS(data) \
					    -attrs [list type submit] \
					    -subelements $fieldtags]] \
	-to $to \
	-command [list negotiate::recv_request_response $xlib $to $i]

    vwait [namespace current]::tmp($i)
    set res $tmp($i)
    unset tmp($i)
    return $res
}

proc negotiate::recv_request_response {xlib jid seq res child} {
    variable tmp

    if {$res != "ok"} {
	set tmp($seq) [list $res $child]
	return
    }

    ::xmpp::xml::split $child tag xmlns attrs cdata subels

    if {$tag == "feature"} {
	::xmpp::xml::split \
	    [lindex $children 0] stag sxmlns sattrs scdata ssubels

	data::draw_window $ssubels \
	    [list [namespace current]::send_negotiation_form $xlib $jid]
    }

    set tmp($seq) [list ok {}]
}

proc negotiate::send_negotiation_form {xlib jid w restags} {
    catch { destroy $w.error.msg }
    $w.bbox itemconfigure 0 -state disabled

    ::xmpp::sendIQ $xlib set \
	-query [::xmpp::xml::create feature \
			-xmlns $::NS(negotiate) \
			-subelement [::xmpp::xml::create x \
					-xmlns $::NS(data) \
					-attrs [list type submit] \
					-subelements $restags]] \
	-to $jid \
	-command [list data::test_error_res $w]
}

